import os
import sys
import argparse
import django
from rested import commands


# main
def main():
    # set environment variable if we are running tests
    if len(sys.argv) >= 2 and sys.argv[1] == 'test':
        os.environ['RESTED_TESTING'] = 'yes'

    # setup django
    sys.path.append(os.getcwd())
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')
    try:
        django.setup()
    except ModuleNotFoundError:
        pass

    # arguments
    parser = argparse.ArgumentParser(description='')
    subparsers = parser.add_subparsers(help='')

    # commands.register('create', commands.create, subparsers)
    commands.register('serve', commands.serve.Command(), subparsers)
    commands.register('shell', commands.shell.Command(), subparsers)
    commands.register('test', commands.test.Command(), subparsers)
    commands.register('manage', commands.manage.Command(), subparsers)
    commands.register('create', commands.create.Command(), subparsers)
    commands.register('worker', commands.worker.Command(), subparsers)
    commands.register('version', commands.version.Command(), subparsers)
    # # commands.register('sockets', commands.create, subparsers)
    # # commands.register('pipeline', commands.create, subparsers)

    try:
        # hijack the command line and run django's default manage.py script if manage is the second argument
        if commands.manage.inject(): return

        # parse args and execute command
        args = parser.parse_args()
        if hasattr(args, 'cmd'):
            sys.exit(args.cmd(args))
        else:
            parser.print_help()
    except ModuleNotFoundError as exc:
        if "No module named 'settings'" not in str(exc): raise exc
        print('Warning: Unable to detect rested/django project in current directory. Failed to load settings.')



from functools import wraps
from rested import abort
from .validators.defined import defined
from .acceptors.undefined import undefined
from .utils import *


# class to build validation rules
class Validator():

    def __init__(self, chain):
        self.chain = chain

    def link(self, validator):
        return self.chain + [passthrough(validator)]

    @classmethod
    def attach(cls, validator):
        # if this takes no arguments attach it as property
        if hasattr(validator, 'dynamic'):
            def wrapped(self, *args, **kwargs):
                return Validator(self.link(validator(*args, **kwargs)))
        # otherwise attach as a function
        else:
            @property
            def wrapped(self):
                return Validator(self.link(validator))
        # attach validator to class
        setattr(cls, validator.__name__, wrapped)

    def execute(self, value):
        for validator in self.chain:
            value = validator(value, accept, reject)
        return value

    @property
    def to(self):
        return Converter(self.chain)

    @property
    def am(self):
        return Validator(self.chain)

    @property
    def accepts(self):
        return Acceptor(self.chain)

# like validator class, but instead of rejecting values short circuits and accepts values
class Acceptor(Validator):

    def link(self, validator):
        return self.chain + [validator]

# like validator class, but is able to modify values
class Converter(Validator):

    def link(self, validator):
        return self.chain + [validator]

# root validator checks if things are even defined before validating or converting
class Root():

    @property
    def to(self):
        return Converter([passthrough(defined)])

    @property
    def am(self):
        return Validator([passthrough(defined)])

    @property
    def accepts(self):
        return Acceptor([passthrough(defined)])

    @property
    def optional(self):
        return Validator([passthrough(undefined)])

# define entry point for building a validator
v = Root()

# attach all of the validators and converters to classes
attach(Validator, "validators", "rested.validation.validators")
attach(Converter, "converters", "rested.validation.converters")
attach(Acceptor, "acceptors", "rested.validation.acceptors")

# check data against schema
def check_recursive(data, schema, errors, cleaned, path):
    current = dig(schema, path)
    if isinstance(current, dict):
        for key in current:
            check_recursive(data, schema, errors, cleaned, path + [key])
    else:
        try:
            value = dig(data, path)
        except KeyError:
            value = Undefined
        validator = dig(schema, path)
        try:
            value = validator.execute(value)
            place(cleaned, path, value)
        except ShortCircuit as acc:
            if acc.value != Undefined:
                place(cleaned, path, acc.value)
        except ValidationError as exc:
            place(errors, path, exc.code)

# check data against schema
def check(data, schema):
    errors = {}
    cleaned = {}
    check_recursive(data, schema, errors, cleaned, [])
    return errors, cleaned

# decorator that checks that a requests data is valid
def validate(schema, field='data'):
    def decorator(view_func):
        @wraps(view_func)
        def wrapped(request, *args, **kwargs):
            errors, data = check(getattr(request, field), schema)
            if errors: abort(400, errors)
            setattr(request, field, data)
            return view_func(request, *args, **kwargs)
        return wrapped
    return decorator



import re
email_regex = re.compile(r"[^@]+@[^@]+\.[^@]+")

def email(v, accept=None, reject=None):
    if not isinstance(v, str): reject('expected_email')
    if not email_regex.match(v): reject('expected_email')



from . import signals
from . import errors
from django.http import JsonResponse

urlpatterns = []
handler404 = errors.handler(404, 'Not Found')
handler500 = errors.handler(500, 'Server Error')
handler403 = errors.handler(403, 'Permission Denied')
handler400 = errors.handler(400, 'Bad Request')
handler_csrf = errors.handler(403, 'Permission Denied: CSRF Failure')

websocketpatterns = signals.patterns



import sys
from subprocess import call
from .command import BaseCommand


class Command(BaseCommand):

    help = 'Run background task worker'

    def add_arguments(self, parser):
        parser.add_argument('-b', '--beats', action='store_true', help='Run beats sheduler with worker')

    def execute(self, args):
        print(args)
        arguments = ''
        if args.beats: arguments = '-B'
        try:
            sys.exit(call(f'celery -A rested.worker worker {arguments} --loglevel=info', shell=True))
        except KeyboardInterrupt:
            pass # should pass these interruprs to subprocess, right now parent closes before subprocess



# chat/consumers.py
from channels.generic.websocket import JsonWebsocketConsumer
from asgiref.sync import async_to_sync
from channels.http import AsgiRequest
from channels.exceptions import DenyConnection
from django.utils.module_loading import import_string


class SignalConsumer(JsonWebsocketConsumer):

    def __init__(self, *args, **kwargs):
        super().__init(*args, **kwargs)
        self.session = None
        self.user = None

    def secure(self):
        # use django middleware to get session and authenticate on initial ws connection
        django_middlewhere = [
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            # need to check csrf but unable to set that in the http request initiating ws, probably set it in query
        ]

        # build a django request from ws request/scope
        self.scope['method'] = 'WEBSOCKET'
        request = AsgiRequest(self.scope, '')

        # get channel's django middleware
        middleware = [import_string(m)(lambda x: None) for m in django_middlewhere]

        # make sure ws request passes channel's django middleware
        for m in middleware:
            if m.process_request(request): raise DenyConnection()

        # if session or user are set store on consumer
        if hasattr(request, 'session'): self.session = request.session
        if hasattr(request, 'user'): self.user = request.user

    def connect(self):
        self.secure()
        self.accept()
        async_to_sync(self.channel_layer.group_add)('signals', self.channel_name)

    def disconnect(self, _):
        async_to_sync(self.channel_layer.group_discard)('signals', self.channel_name)

    def forward(self, event):
        # could reload user from session and check is_authenticated
        self.send_json({'message': event['data']})

    def receive_json(self, data):
        message = data['message']
        self.send_json({'message': message})



import os
import shutil
from .command import BaseCommand


class Command(BaseCommand):

    help = 'Create an empty rested django project'

    def add_arguments(self, parser):
        parser.add_argument('name', type=str, help='project name / directory to create')

    def execute(self, args):
        working_dir = os.getcwd()
        file_dir = os.path.dirname(os.path.abspath(__file__))
        destination = os.path.join(working_dir, args.name)
        source = os.path.normpath(os.path.join(file_dir, '../templates/default'))
        try:
            shutil.copytree(source, destination)
            print(f"Project {args.name} successfully created at {destination}")
        except FileExistsError:
            print(f"Unable to create project, the specified path '{destination}' already exists.")



from functools import wraps


def passes_test(test_func, code, message):
    def decorator(view_func):
        @wraps(view_func)
        def wrapped(request, *args, **kwargs):
            if test_func(request):
                return view_func(request, *args, **kwargs)
            return (code, {'error': message})
        return wrapped
    return decorator


def public(func):
    return func

def authenticated(func):
    message = 'User must be authenticated to access this resource'
    return passes_test(lambda r: r.user.is_authenticated, 401, message)(func)

def admin(func):
    message = 'You do not have permission to access this resource'
    return authenticated(passes_test(lambda r: r.user.is_superuser, 403, message)(func))

def staff(func):
    message = 'You do not have permission to access this resource'
    return authenticated(passes_test(lambda r: r.user.is_staff or r.user.is_superuser, 403, message)(func))



from celery import Celery
from celery.schedules import crontab # import for convience


# create celery worker app
app = Celery('rested')
app.config_from_object('django.conf:settings', namespace='CELERY')

# create a background task
task = app.task

# create a scheduled task
def scheduled_task(time, args=None, soft_time_limit=None, time_limit=None):

    # check time limit arguments
    if time_limit:
        assert soft_time_limit < time_limit, "soft_time_limit should be less than time_limit"
        assert time_limit < time.total_seconds(), "time_limit should be less than 'time', the call period"

    # create decorator
    def decorator(func):

        # add func as task
        tsk = app.task(func)
        if time_limit or soft_time_limit:
            tsk.soft_time_limit = soft_time_limit
            tsk.time_limit = time_limit

        # add task to schedule
        label = func.__module__ + '.' + func.__name__
        schd = app.conf.CELERYBEAT_SCHEDULE
        schd[label] = {'task': label, 'schedule': time, 'args': args}
                                        # removing for now, but keeping comment in case there was some reason
        # update runner's schedule
        app.conf.update(CELERYBEAT_SCHEDULE=schd)
        return func
    return decorator



# Create your tasks here
from __future__ import absolute_import, unicode_literals
from rested.worker import task, scheduled_task


@scheduled_task(1)
def hello():
    print('hello')

@scheduled_task(2)
def goodbye():
    print('goodbye')

@task
def add(x, y):
    print('celery', x + y)
    return x + y



import os
import importlib
from functools import wraps
from inspect import getmembers, isfunction


# class to represent a value that was undefined, this lets us differentiate between a passed null and non existant
class Undefined(): pass

# error class used to short circuit validation chain and accept a value
class ShortCircuit(Exception):

    def __init__(self, value):
        self.value = value
        super().__init__()

# validation error called when validation is rejected
class ValidationError(Exception):

    def __init__(self, code):
        self.code = code
        super().__init__(code)

# utility function to accept value
def accept(value):
    raise ShortCircuit(value)

# utility function to reject value
def reject(code):
    raise ValidationError(code)

# call validation function then return value
def passthrough(func):
    @wraps(func)
    def wrapped(value, *args, **kwargs):
        func(value, *args, **kwargs)
        return value
    return wrapped

# get a value in a nested dictionary
def dig(dictionary, keys):
    value = dictionary
    if not isinstance(keys, list):
        keys = [keys]
    for k in keys:
        if not isinstance(value, dict): raise KeyError(k)
        value = value[k]
    return value

# put a value in a nested dictionary, created dicts along path when missing
def place(dictionary, keys, value):
    for key in keys[:-1]:
        dictionary = dictionary.setdefault(key, {})
    dictionary[keys[-1]] = value

# magic function to import and attach all validation functions from the specified directory
def attach(cls, dpth, dot):
    vdir = os.path.dirname(os.path.realpath(__file__)) + '/' + dpth
    tree = os.listdir(vdir)
    for path in tree:
        if path.startswith('_'): continue
        module = importlib.import_module(dot + "." + path[:-3])
        validators = getmembers(module, isfunction)
        for _, validator in validators:
            cls.attach(validator)


# check for extra keys in data given schema
# def extras(data, schema):
#     extra = set(data.keys()) - set(schema.keys())
#     for k in schema:
#         if type(schema.get(k)) is dict and type(data.get(k)) is dict:
#             extra.append(extras(data[k], schema[k]))
#     return extra



import pytest


def database(transactional=False, reset_sequences=False, autouse=True):
    from pytest_django.fixtures import _django_db_fixture_helper as helper
    @pytest.fixture(autouse=autouse)
    def db(request, django_db_setup, django_db_blocker):
        if reset_sequences: assert transactional, "Reset sequences only works when transactional is True."
        helper(request, django_db_blocker, transactional=transactional, reset_sequences=reset_sequences)
    return db

@pytest.fixture()
def rest():
    from rested.test import Client
    return Client()



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer


def signal(user, data):
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)(f'signals.{user.id}', {'type': 'forward', 'data': data})



import pytest
import unittest
from rested.test.fixtures import database
from django.test import Client
from api.models import Question


db = database(reset_sequences=False, autouse=True)

#@pytest.mark.django_db
class SimpleTest(unittest.TestCase):
    def setUp(self):
        # Every test needs a client.
        self.client = Client()

    def test_details(self):
        # Issue a GET request.
        response = self.client.get('/')

        # Check that the response is 200 OK.
        self.assertEqual(response.status_code, 302)

    def test_sequence(self):
        q = Question.create().save()
        assert q.id == 1

    def test_sequence_again(self):
        q = Question.create().save()
        assert q.id == 1



from functools import partial, reduce
from django.urls.conf import _path
from django.urls.resolvers import RoutePattern, RegexPattern
import re


# factory method that creates a pattern and tags it with a method
def factory(pattern, method):
    def wrapper(*args, **kwargs):
        p = pattern(*args, **kwargs)
        p.method = method
        return p
    return wrapper

# create a rest route scoped by method
def rest(route, method, view, decorators=None, kwargs=None, name=None, pattern='route'):

    # accept a decorator or list of decorators
    if not decorators: decorators = []
    if not isinstance(decorators, (list, tuple)): decorators = [decorators]

    # wrap view with decorators, decorators are applied right to left (bottom to top)
    for decorator in reversed(decorators):
        view = decorator(view)

    # proxy path function
    Pattern = RoutePattern if pattern == 'route' else RegexPattern
    return _path(route, view, kwargs=kwargs, name=name, Pattern=factory(Pattern, method))



import json as sjson
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.fields.files import FieldFile

class RestedJSONEncoder(DjangoJSONEncoder):

    def default(self, o):
        if isinstance(o, FieldFile):
            try:
                return o.url
            except ValueError:
                return ''
        return super().default(o)

def dumps(*args, **kwargs):
    kwargs['cls'] = kwargs.get('cls', RestedJSONEncoder)
    return sjson.dumps(*args, **kwargs)

def loads(*args, **kwargs):
    return sjson.loads(*args, **kwargs)



from urllib.parse import urlencode
from rested import json
from django.test import Client


class RestClient(Client):

    def get(self, url, params=None):
        params = '?' + urlencode(params) if params else ''
        response = super().get(url + params)
        response.data = response.json()
        response.status = response.status_code
        return response

    def put(self, url, data=None, params=None, json_content=True):
        params = '?' + urlencode(params) if params else ''
        if json_content:
            response = super().put(url + params, json.dumps(data), content_type="application/json")
        else:
            response = super().put(url + params, data)
        response.data = response.json()
        response.status = response.status_code
        return response

    def post(self, url, data=None, params=None, json_content=True):
        params = '?' + urlencode(params) if params else ''
        if json_content:
            response = super().post(url + params, json.dumps(data), content_type="application/json")
        else:
            response = super().post(url + params, data)
        response.data = response.json()
        response.status = response.status_code
        return response

    def delete(self, url, params=None):
        params = '?' + urlencode(params) if params else ''
        response = super().delete(url + params)
        response.data = response.json()
        response.status = response.status_code
        return response



# Generated by Django 3.0 on 2020-01-19 04:59

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='choice',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='choice',
            name='updated_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='question',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='question',
            name='updated_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



"""
Django settings for template project.

Generated by 'django-admin startproject' using Django 3.0.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

# Set environment
ENVIRONMENT = os.environ.get('ENVIRONMENT', 'production').lower()

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.abspath(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'replacethisfakesecret'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'channels',
    'api',
    'django_extensions',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'rested.middleware.RestMiddleware',
]

ROOT_URLCONF = 'api.urls'
ASGI_APPLICATION = 'asgi.application'


# Database
# https://docs.djangoproject.com/en/3.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Channels
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [("redis", 6379)],
        },
    },
}

# Tasks
# CELERY_TASK_ALWAYS_EAGER=True
CELERY_BROKER_URL = 'redis://redis:6379'
CELERY_RESULT_BACKEND = CELERY_BROKER_URL

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = '/static/'



import sys
import pytest
from subprocess import call
from .command import BaseCommand
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from rested.test import plugin


class TestEventHandler(PatternMatchingEventHandler):
    def __init__(self):
        super().__init__(ignore_directories=True, ignore_patterns=[
            '*.pytest_cache*',
            '*__pycache__*',
            '*.cache',
            '*.sqlite3',
            '*.pyc'])
        self.changed = True

    def check(self, handler):
        if self.changed:
            self.changed = False
            handler()

    def on_any_event(self, event):
        # print(event)
        self.changed = True

def watch(on_change, path='.'):
    path = '.'
    handler = TestEventHandler()
    observer = Observer()
    observer.schedule(handler, path, recursive=True)
    observer.start()
    try:
        while observer.isAlive():
            handler.check(on_change)
            observer.join(0.25)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


class Command(BaseCommand):

    help = 'Discover and run tests in the specified modules or the current directory'

    def add_arguments(self, parser):
        parser.add_argument('pattern', type=str, nargs='?', help='all tests matching this pattern will be run')
        parser.add_argument('-f', '--failfast', action='store_true', help='stop tests after first failure')
        parser.add_argument('-k', '--keepdb', action='store_true', help='keep db between tests')
        parser.add_argument('-w', '--watch', action='store_true', help='rerun tests when files change')

    def execute(self, args):
        # setup tests
        argv = []
        if args.pattern:
            argv += ['-k', args.pattern]
        if args.failfast:
            argv.append('--exitfirst')
        if args.keepdb:
            argv.append('--reuse-db')

        # run tests (with or without watcher)
        if args.watch:
            return watch(lambda: call([arg for arg in sys.argv if arg not in ['-w', '--watch']]))
        return pytest.main(argv, plugins=[plugin])



# register a command
def register(name, command, subparsers):
    parser = subparsers.add_parser(name, help=command.help)
    parser.set_defaults(cmd=command.execute)
    command.add_arguments(parser)

# rested command patterned after django command
class BaseCommand:
    help = ''

    def add_arguments(self, parser):
        pass

    def execute(self, args):
        pass

# mixin that enables us to execute django commands from rested command line
class DjangoCommand:
    def execute(self, args):
        options = vars(args)
        args = options.pop('args', ())
        base_options = {'verbosity': 1, 'force_color': False, 'no_color': False, 'skip_checks': False}
        super().execute(*args, **{**options, **base_options})



import pytest
from rested.test.fixtures import database
# pytestmark = pytest.mark.django_db


db = database(reset_sequences=False, autouse=True)

def test_ping(rest):
    response = rest.get('/ping')
    assert response.status == 200
    assert response.data == {"data": "pong"}



def max(limit):
    def validator(v, accept=None, reject=None):
        try:
            v = len(v)
        except TypeError:
            pass
        if v > limit: reject(f'expected_max_{limit}')
    return validator
max.dynamic = True



# Generated by Django 3.0 on 2020-01-05 07:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Question',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('question_text', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='Choice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('choice_text', models.CharField(max_length=200)),
                ('votes', models.IntegerField(default=0)),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.Question')),
            ],
        ),
    ]



import re


# strong password means: 8 characters, one capital, one number, one lowercase
digits = re.compile('[0-9]')
lowercase = re.compile('[a-z]')
uppercase = re.compile('[A-Z]')

def password(v, accept=None, reject=None):
    if not isinstance(v, str): reject('expected_strong_password')
    strong = digits.search(v) and lowercase.search(v) and uppercase.search(v) and len(v) >= 8
    if not strong : reject('expected_strong_password')



from django_extensions.management.commands.shell_plus import Command as ShellPlusCommand
from .command import DjangoCommand

class Command(DjangoCommand, ShellPlusCommand):
    help = 'Interactive python shell with all models autoloaded'



from django.shortcuts import redirect
import django.contrib.auth as auth
from django.contrib.auth.models import User
from django.views.decorators.csrf import csrf_exempt
from rested import signal
from . import tasks


@csrf_exempt
def index(request):
    print(User.objects.all())
    return redirect('/static/index.html')

@csrf_exempt
def ping(request):
    tasks.add.delay(1, 2)
    return {'data': 'pong'}

@csrf_exempt
def login(request):
    user = User.objects.filter(username='john').first()
    print(user)
    auth.login(request, user)
    return {}

@csrf_exempt
def logout(request):
    auth.logout(request)
    return {}

@csrf_exempt
def message(request):
    signal(request.user, request.data['message'])
    return {'message': request.data['message']}



from channels.generic.websocket import JsonWebsocketConsumer
from asgiref.sync import async_to_sync
from channels.http import AsgiRequest
from channels.exceptions import DenyConnection
from django.contrib.auth import get_user
from django.utils.module_loading import import_string
from django.contrib.auth.signals import user_logged_out
from django.dispatch import receiver
from channels.layers import get_channel_layer


# websocket signal consumer
class SignalConsumer(JsonWebsocketConsumer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = None
        self.user = None

    def secure(self):
        # use django middleware to get session and authenticate on initial ws connection
        django_middlewhere = [
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            # need to check csrf but unable to set that in the http request initiating ws, probably set it in query
        ]

        # build a django request from ws request/scope
        self.scope['method'] = 'WEBSOCKET'
        request = AsgiRequest(self.scope, '')

        # get channel's django middleware
        middleware = [import_string(m)(lambda x: None) for m in django_middlewhere]

        # make sure ws request passes channel's django middleware
        for m in middleware:
            if m.process_request(request): raise DenyConnection()

        # set session and user
        if hasattr(request, 'session'): self.session = request.session
        if hasattr(request, 'user'): self.user = request.user

        # deny if we don't have a session and an authenticated user
        if not self.session or not self.user or not self.user.is_authenticated:
            raise DenyConnection()

    def connect(self):
        self.secure()
        self.accept()
        async_to_sync(self.channel_layer.group_add)(f'signals.{self.user.id}', self.channel_name)

    def disconnect(self, _):
        async_to_sync(self.channel_layer.group_discard)(f'signals.{self.user.id}', self.channel_name)

    # when data is sent a heart beat request respond with empty object
    def receive_json(self, _):
        self.send_json({})

    def forward(self, event):
        # could reload user from session and check is_authenticated
        self.send_json({'data': event['data']})

    def check_session(self, data):
        # could also check session still exists in store self.session.load()
        if data['session_key'] == self.session.session_key:
            self.disconnect('Expired Session')


# when a user logged out disconnect websockets with the now invalidated session
@receiver(user_logged_out)
def on_user_logged_out(sender, request, user, **kwargs):
    channel_layer = get_channel_layer()
    if not user or not request.session: return
    session_key = request.session.session_key
    async_to_sync(channel_layer.group_send)(f'signals.{user.id}', {
        'type': 'check_session',
        'session_key': session_key
    })



from django.http import JsonResponse


class AbortException(Exception):
    def __init__(self, status, data):
        self.status = status
        self.data = data

def abort(status, data=None):
    if not data: data = {}
    raise AbortException(status, data)

def handler(status, message, data=None):
    def handler(request, exception=None, reason=None):
        return JsonResponse(data={
            'status': status,
            'message': message,
            'data': data if data is not None else {}
        }, status=status)
    return handler



from channels.management.commands.runserver import Command as RunServerCommand
from .command import DjangoCommand

class Command(DjangoCommand, RunServerCommand):
    default_addr = '0.0.0.0'
    default_addr_ipv6 = '::'



from django.db.models import Model as DjangoModel, DateTimeField
from django.forms.models import model_to_dict
from django.utils import timezone
from django.db.models import * # so we can rested import models


class Model(DjangoModel):

    created_at = DateTimeField(default=timezone.now)
    updated_at = DateTimeField(default=timezone.now)

    class Meta:
        abstract = True

    @classmethod
    def create(cls, values=None, **kwargs):
        values = values or kwargs
        return cls.objects.create(**values)

    def update(self, values=None, **kwargs):
        values = values or kwargs
        for field in list(values.keys()):
            setattr(self, field, values[field])
        return self

    def to_dict(self, *args, **kwargs):
        return model_to_dict(self, *args, **kwargs)

    def save(self, *args, **kwargs):
        self.updated_at = timezone.now()
        super().save(*args, **kwargs)
        return self



from rested import models


class Question(models.Model):
    question_text = models.CharField(max_length=200)


class Choice(models.Model):
    question = models.ForeignKey(Question, on_delete=models.CASCADE)
    choice_text = models.CharField(max_length=200)
    votes = models.IntegerField(default=0)



import os
from .command import BaseCommand


class Command(BaseCommand):

    help = 'Display rested version'

    def execute(self, args):
        file_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(file_dir, '../version.txt')) as file:
            print(file.read().strip())



from django.core.exceptions import ObjectDoesNotExist


# converts a field into an orm model by looking on specified field, validation error if record doesn't exist
# usage: v.am.string.to.model
def model(model, field='id'):
    def validator(v, accept=None, reject=None):
        try:
            fields = {}
            fields[field] = v
            return model.objects.get(**fields)
        except ObjectDoesNotExist:
            reject(f'expected_model_exists')
    return validator
model.dynamic = True



from django.urls import path
from rested import rest, authorize
from rested import signals
from . import views


urlpatterns = [
    path('', views.index, name='index'),
]

restpatterns = [
    rest("ping", "get", views.ping),
    rest("login", "post", views.login),
    rest("logout", "post", views.logout),
    rest("message", "put", views.message, authorize.authenticated),
]

websocketpatterns = signals.patterns



from django.test import TestCase
from api.models import Question


class AnimalTestCase(TestCase):
    def setUp(self):
        Question.objects.create(question_text="lion")
        Question.objects.create(question_text="cat")

    def test_animals_can_speak(self):
        self.assertEqual(Question.objects.count(), 2)



# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

# use pytest for test command
class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = "--ignore=rested"
    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)

# package settings
setup(

    # info
    name='django-rested',
    version=open('rested/version.txt').read().strip(),
    author='Stephen Quebe',
    author_email='stephen@mochi.ai',
    url='https://github.com/mochi-ai/rested.git',
    description='Make creating rest APIs in Django simple.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",

    # include
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,

    # public dependencies
    install_requires=[
        'django==3.0',
        'channels==2.4.0',
        'channels-redis==2.4.1',
        'django-extensions==2.2.6',
        'ipython==7.11.1',
        'pytest==5.3.1',
        'pytest-django==3.7.0',
        'pytest-mock==3.2.0',
        'watchdog==0.9.0',
        'celery==4.4.0',
        'redis==3.3.11'],

    # cli scripts
    entry_points={'console_scripts': ['rested = rested.cli:main']},

    # tests
    tests_require=['django', 'pytest', 'pytest-django', 'pytest-watch', 'pytest-mock'],
    cmdclass={"test": PyTest},
)



import os
import sys
import django.__main__
from django.core.management import execute_from_command_line
from .command import BaseCommand


class Command(BaseCommand):
    help = 'Run a django management command'

def inject():
    if len(sys.argv) >= 2 and sys.argv[1] == 'manage':
        # trim 'rested' off arguments and set to simulate "python -m django" for management and runserver autoreload
        sys.argv = sys.argv[1:]
        sys.argv[0] = django.__main__.__file__

        # run django's original manage.py command
        execute_from_command_line(sys.argv)
        return True



import os
import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
django.setup()

from api import urls
from channels.routing import ProtocolTypeRouter, URLRouter

application = ProtocolTypeRouter({
    # (http->django views is added by default)
    'websocket': URLRouter(urls.websocketpatterns),
})



from rested import json
from importlib import import_module
from django.conf import settings
from django.http import JsonResponse, HttpResponse
from django.db.models.query import QuerySet
from functools import lru_cache
from .errors import AbortException


class RestURLPattern:
    def __init__(self, method):
        # get default configuration module
        urlconf = import_module(settings.ROOT_URLCONF)

        # clone url config
        for x in dir(urlconf):
            setattr(self, x, getattr(urlconf, x))

        # copy url patterns and rest patterns that match method (create method scoped urlconf)
        regular_patterns = getattr(urlconf, "urlpatterns", [])
        rest_patterns = [p for p in getattr(urlconf, "restpatterns", []) if p.pattern.method == method.lower()]
        self.urlpatterns = regular_patterns + rest_patterns


@lru_cache(maxsize=None)
def scoped_url_patterns(method):
    return RestURLPattern(method)

def attach_data(request):
    if 'application/json' in request.META.get('CONTENT_TYPE', []):
        try:
            return json.loads(request.body.decode())
        except json.sjson.decoder.JSONDecodeError:
            return {}
    else:
        return {k:request.POST[k] for k in request.POST}

def attach_params(request):
    return {k:request.GET[k] for k in request.GET}

def wrap_response(response):
    # if response is a tuple, then someone is return (status, data), otherwise assume data and implied 200 status
    if isinstance(response, tuple):
        status, data = response
    else:
        status, data = (200, response)

    # convert data and status into json response
    if isinstance(data, HttpResponse):
        return data
    if isinstance(data, dict):
        return JsonResponse(data, status=status, encoder=json.RestedJSONEncoder)
    if hasattr(data, 'to_dict'):
        return JsonResponse(data.to_dict(), status=status, encoder=json.RestedJSONEncoder)
    if isinstance(data, QuerySet):
        return JsonResponse({m.id: m.to_dict() for m in data}, status=status, encoder=json.RestedJSONEncoder)
    return JsonResponse(data, status=status, encoder=json.RestedJSONEncoder)


class RestMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.data = attach_data(request)
        request.params = attach_params(request)
        request.urlconf = scoped_url_patterns(request.method)
        response = self.get_response(request)
        return wrap_response(response)

    def process_exception(self, request, exception):
        if type(exception) is AbortException:
            return (exception.status, exception.data)
        return None
